module Ruby

    # 
    # Ruby::Stack
    #
    # Implements a simple call stack using Ruby::Frame.metadata.
    #
    # Allows for methods called below to see values put
    # on the stack by methods higher in the call stack.
    #
    # Useful as a replacement for Thread.current[:__secret] style
    # programming.
    #
    class Stack

        # 
        # Access the value associated with +key+ on the stack.
        # Returns nil if there is no value.
        # 
        def self.[](key)
            caller.frames.each do |f|
                if f.metadata[self]
                    v = f.metadata[self][key]
                    return v if v
                end
            end

            nil
        end

        #
        # Pushes a value +val+ for +key+ on the current stack.
        # All methods called from this method can access the values
        # using Ruby::Stack.[].
        # 
        def self.[]=(key,val)
            fr = caller.frames.first
            fr.metadata[self] ||= Hash.new
            fr.metadata[self][key] = val
            val
        end
    end
end

if $0 == __FILE__
    def a
        Ruby::Stack[:age] = 12
        yield
    end

    def b
        puts "age is #{Ruby::Stack[:age]}"
    end

    a { b }

    p Ruby::Stack[:age]

end
