
#define __USE_UNIX98

#include "ruby.h"
#include "runtime.h"
#include "intern.h"
#include "internal.h"
#include "state.h"

#include <errno.h>
/*
static VALUE rb_osthread_group_init(VALUE self) {
    rb_iv_set(self, "@threads", rb_ary_new());
    return self;
}

static VALUE rb_osthread_group_add(VALUE self, VALUE th) {
    VALUE ary;
    ary = rb_iv_get(self, "@threads");
    rb_ary_push(ary, th);
    return th;
}
*/

struct rb_os_thread_stub {
    VALUE thread;
    ruby_state *state;
    VALUE block;
    VALUE arg;
};

VALUE rb_osthread_init(VALUE self) {
    return Qnil;
}
/*
static void *osthread_term(int sig) {
    printf("is osthread_term...\n");
    pthread_exit(NULL);
}
*/

int rb_osthread_is_me(VALUE th) {
    return pthread_equal(pthread_self(), GET_THREAD(th)->os_thread);
}

void rb_osthread_lock(VALUE self) {
    pthread_mutex_t *m;
    struct rb_os_thread_t *oth;
    if(self == (VALUE)0) {
        oth = GET_THREAD(STATE_VAR(thread));
    } else {
        oth = GET_THREAD(self);
    }
    m = &(oth->state->xcritical);
    pthread_mutex_lock(m);
}

void rb_osthread_unlock(VALUE self) {
    pthread_mutex_t *m;
    struct rb_os_thread_t *oth;
    if(self == (VALUE)0) {
        oth = GET_THREAD(STATE_VAR(thread));
    } else {
        oth = GET_THREAD(self);
    }
    m = &(oth->state->xcritical);
    pthread_mutex_unlock(m);
}

void rb_osthread_signal(VALUE self, int sig) {
    pthread_kill(GET_THREAD(self)->os_thread, sig);
}

#define SIG_RBOP 30

void rb_gc_mark_frame_list(struct FRAME *);
void rb_gc_mark_stack(void*, void*);

void rb_osthread_mark_threads() {
    VALUE ary, thread;
    struct rb_os_thread_t *oth;
    int i = 0;

    ary = STATE_VAR(os_threads);
    // printf("marking %lu threads in %lu\n", RARRAY(ary)->len, STATE_VAR(main_osthread));
    for(i = 0; i < RARRAY(ary)->len; i++) {
        thread = RARRAY(ary)->ptr[i];
        oth = GET_THREAD(thread);
        
        if(oth->locals) {
            rb_gc_mark(oth->locals);
        }
        if(oth->error) {
            rb_gc_mark(oth->error);
        }
        rb_gc_mark_frame_list(oth->state->frames);

        //printf("locking and marking %lu\n", thread);

        if(!rb_osthread_is_me(thread)) {
            GET_THREAD(thread)->status = THREAD_GC;
            rb_osthread_signal(thread, SIG_RBOP);
        }
    }
    // printf("done marking threads!\n");
}

void osthread_wait_for(VALUE thread, int status) {
    while(GET_THREAD(thread)->status != status) {
        sched_yield();
    }
}

void osthread_until_running(VALUE thread) {
    osthread_wait_for(thread, THREAD_RUNNING);
}


void osthread_set_status(VALUE thread, int set, int wait) {
    if(set > 0) {
        GET_THREAD(thread)->status = set;
    }
    
    rb_osthread_signal(thread, SIG_RBOP);
    
    if(wait > 0) {
        osthread_wait_for(thread, wait);
    }
}

void rb_osthread_unlock_threads() {
    VALUE ary, th;
    int i;

    ary = STATE_VAR(os_threads);
    for(i = 0; i < RARRAY(ary)->len; i++) {
        th = RARRAY(ary)->ptr[i];
        if(!rb_osthread_is_me(th)) {
            osthread_set_status(th, -1, THREAD_RUNNING);
        }
    }
}

static void *osthread_cleanup(void *d) {
    struct rb_os_thread_t *oth;
    VALUE thread;
    thread = (VALUE)d;
    rb_ary_delete(STATE_VAR(os_threads), thread);
    oth = GET_THREAD(thread);
    oth->status = THREAD_KILLED;

    return NULL;
}

static void *rb_osthread_stub(void *st) {
    int state;
    struct rb_os_thread_t *oth;
    struct rb_os_thread_stub *info;
    VALUE thread;

    /*
    memset(&term, 0, sizeof(term));
    sigemptyset(&term.sa_mask);
    term.sa_flags = 0;
    term.sa_handler = osthread_term;
    */
    info = (struct rb_os_thread_stub*)st;

    inherit_from_state(info->state);

    STATE_VAR(thread) = thread = info->thread;
    rb_ary_push(STATE_VAR(os_threads), thread);

    Data_Get_Struct(thread, struct rb_os_thread_t, oth);

    oth = GET_THREAD(thread);
    oth->error = 0;
    oth->state = rb_get_state();

    pthread_cleanup_push((void *)osthread_cleanup, (void*)thread);
    pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
    oth->status = THREAD_RUNNING;
    PUSH_FRAME();
    PUSH_SCOPE();
    ruby_scope->local_vars = 0;
    ruby_scope->local_tbl  = 0;
    top_scope = ruby_scope;
    
    PUSH_TAG(PROT_NONE);

    RUN_TAG(state) {
        if(info->arg) {
            oth->result = rb_funcall(info->block, _S(call), 1, info->arg);
        } else {
            oth->result = rb_funcall(info->block, _S(call), 0);
        }
    }

    POP_TAG();
    POP_SCOPE();
    POP_FRAME();
    pthread_cleanup_pop(1);

    if(state) {
        oth->error = ruby_errinfo;
    }
    
    return NULL;
}

/*
 * call-seq: Thread::OS.new([arg]) { |arg| }
 *
 * Start a new OS thread which begins by executing the given block.
 * +arg+ is optional and is passed into the block as the only argument
 * if given, just like +Thread+.
 */
static VALUE rb_osthread_start(int argc, VALUE *argv, VALUE self)
{
    struct rb_os_thread_t *oth;
    struct rb_os_thread_stub *info;
    VALUE thread;

    if(!rb_block_given_p()) {
        rb_raise(CPATH(ArgumentError), "Block required");
    }
    
    thread = Data_Make_Struct(CPATH(Thread::OS), struct rb_os_thread_t, 0, 0, oth);

    info = malloc(sizeof(struct rb_os_thread_stub));
    if(argc == 1) {
        info->arg = argv[0];
    } else {
        info->arg = 0;
    }
    info->thread = thread;
    info->state = rb_get_state();
    info->block = rb_block_proc();
    
    oth->locals = rb_hash_new();
    oth->status = 0;

    rb_obj_call_init(thread, 0, NULL);

    pthread_create(&oth->os_thread, NULL, rb_osthread_stub, info);
    
    // Wait til the thread is started.
    while(!oth->status) {
        sched_yield();
    }

    return thread;
}

/*
 * call-seq: Thread::OS.thread_id
 *
 * Returns an integer that is a unique ID per thread.
 */
VALUE rb_osthread_id(VALUE self) {
    return UINT2NUM(pthread_self());
}

/*
 * call-seq: th[key]
 *
 * Returns a thread local variable for +key+
 */
VALUE rb_osthread_aref(VALUE self, VALUE name) {
    VALUE hash = GET_THREAD(self)->locals;
    return rb_hash_aref(hash, name);
}

/*
 * call-seq: th[key] = val
 *
 * Sets a thread local variable +key+ to +val+
 */
VALUE rb_osthread_aset(VALUE self, VALUE name, VALUE val) {
    VALUE hash = GET_THREAD(self)->locals;
    return rb_hash_aset(hash, name, val);
}

/*
 * call-seq: Thread::OS.current
 *
 * Returns the current Thread::OS.
 */
VALUE rb_osthread_current(VALUE self) {
    return STATE_VAR(thread);
}

/*
 * call-seq: th.join => val
 *
 * Blocks the calling thread until +th+ is finished. If an exception is 
 * raised inside +th+, then exception is raised in the calling thread here.
 *
 * +val+ is the return value of the thread.
 */
VALUE rb_osthread_join(VALUE self) {
    struct rb_os_thread_t *oth;
    int state;
    VALUE err;

    oth = GET_THREAD(self);

    state = pthread_join(oth->os_thread, NULL);

    if(state) {
        switch(state) {
            case ESRCH:
                err = rb_const_get(CPATH(Thread::OS), _S(UnknownThread));
                rb_raise(err, "unknown thread");
            case EINVAL:
                err = rb_const_get(CPATH(Thread::OS), _S(InvalidThread));
                rb_raise(err, "invalid thread");
            case EDEADLK:
                err = rb_const_get(CPATH(Thread::OS), _S(DeadLock));
                rb_raise(err, "would deadlock");
        }
        err = rb_const_get(CPATH(Thread::OS), _S(Error));
        rb_raise(err, "error");
    }

    if(oth->error) {
        rb_exc_raise(oth->error);
    }

    if(oth->result) {
        return oth->result;
    } else {
        return Qnil;
    }
}

/*
 * call-seq: th.detach
 *
 * Detachs +th+ from the rest of the threads. A detached thread
 * can not be join and cleans up it's resources as soon as it
 * finishes.
 */
VALUE rb_osthread_detach(VALUE self) {
    int state;
    VALUE err;

    state = pthread_detach(GET_THREAD(self)->os_thread);

    if(state) {
        switch(state) {
            case ESRCH:
                err = rb_const_get(CPATH(Thread::OS), _S(UnknownThread));
                rb_raise(err, "unknown thread");
            case EINVAL:
                err = rb_const_get(CPATH(Thread::OS), _S(InvalidThread));
                rb_raise(err, "invalid thread");
        }
        err = rb_const_get(CPATH(Thread::OS), _S(Error));
        rb_raise(err, "error");
    }

    return self;
}

/*
 * call-seq: th == th2
 *
 * Compares +th+ to +th2+ and returns true if the match the same
 * OS thread.
 */
VALUE rb_osthread_equal(VALUE self, VALUE other) {
    if(pthread_equal(GET_THREAD(self)->os_thread,
                GET_THREAD(other)->os_thread)) {
        return Qtrue;
    } else {
        return Qfalse;
    }
}

/*
 * call-seq: th.policy => int
 *
 * Returns the current scheduling policy of +th+. See +policy=+ for 
 * the list of all policies.
 */
VALUE rb_osthread_policy(VALUE self) {
    int state, policy;
    struct sched_param param;
    VALUE err;
    state = pthread_getschedparam(GET_THREAD(self)->os_thread,
            &policy, &param);
    if(state) {
        switch(state) {
            case ESRCH:
                err = rb_const_get(CPATH(Thread::OS), _S(UnknownThread));
                rb_raise(err, "unknown/invalid thread");
        }
        err = rb_const_get(CPATH(Thread::OS), _S(Error));
        rb_raise(err, "error");
    }

    return INT2NUM(policy);
}

/*
 * call-seq: th.policy = val
 *
 * Sets the scheduling policy of +th+.
 */
VALUE rb_osthread_set_policy(VALUE self, VALUE policy) {
    int state;
    struct sched_param param = {0};
    VALUE err;

    state = pthread_setschedparam(GET_THREAD(self)->os_thread,
            policy, &param);

    if(state) {
        switch(state) {
            case ESRCH:
                err = rb_const_get(CPATH(Thread::OS), _S(UnknownThread));
                rb_raise(err, "unknown thread");
            case EINVAL:
                err = rb_const_get(CPATH(Thread::OS), _S(InvalidThread));
                rb_raise(err, "invalid thread");
            case EPERM:
                err = rb_const_get(CPATH(Thread::OS), _S(Permission));
                rb_raise(err, "no superuser permissions");
        }
        err = rb_const_get(CPATH(Thread::OS), _S(Error));
        rb_raise(err, "error");
    }

    return Qnil;
}

/*
 * call-seq: Thread::OS.list => ary
 *
 * Returns a list of all Thread::OS instances.
 */
VALUE rb_osthread_list(VALUE self) {
    return STATE_VAR(os_threads);
}

int rb_get_signo(VALUE signal);

/*
 * call-seq: th.kill(sig)
 *
 * Sends +th+ the signal +sig+. See Process#kill for information about
 * signals.
 */
VALUE rb_osthread_kill(VALUE self, VALUE signal) {
    int state, sig;
    VALUE err;
    sig = rb_get_signo(signal);
    if((state = pthread_kill(GET_THREAD(self)->os_thread, sig))) {
        switch(state) {
            case ESRCH:
                err = rb_const_get(CPATH(Thread::OS), _S(UnknownThread));
                rb_raise(err, "unknown thread");
            case EINVAL:
                err = rb_const_get(CPATH(Thread::OS), _S(InvalidThread));
                rb_raise(err, "invalid signal");
        }
        err = rb_const_get(CPATH(Thread::OS), _S(Error));
        rb_raise(err, "error");
    }
    return Qnil;
}

/*
 * call-seq: th.cancel
 *
 * Cancels +th+. It ceases to run as soon as cancel returns. Note that
 * if the thread has not been detached (see Thread::OS#detach), the thread
 * must be joined (Thread::OS#join) to release it's resources.
 */
VALUE rb_osthread_cancel(VALUE self) {
    VALUE err;
    int state;
    if((state = pthread_cancel(GET_THREAD(self)->os_thread))) {
        switch(state) {
            case ESRCH:
                err = rb_const_get(CPATH(Thread::OS), _S(UnknownThread));
                rb_raise(err, "unknown thread");
        }
        err = rb_const_get(CPATH(Thread::OS), _S(Error));
        rb_raise(err, "error");

    }
    osthread_wait_for(self, THREAD_KILLED);
    return Qnil;
}

/*
 * call-seq: th.state
 *
 * Returns the Ruby::State instance that is used inside +th+.
 */
VALUE rb_osthread_state(VALUE self) {
    ruby_state *st;
    VALUE out;

    st = GET_THREAD(self)->state;

    out = Data_Wrap_Struct(CPATH(Ruby::State), 0, 0, st);

    return out;
}

/*
 * call-seq: th.pause
 *
 * Pauses or unpauses +th+. If +th+ is already paused, +th+ is unpaused.
 * While paused, the thread performs no operations.
 */
VALUE rb_osthread_pause(VALUE self) {
    if(GET_THREAD(self)->status == THREAD_RUNNING) {
        // Wait til it's paused.
        osthread_set_status(self, THREAD_PAUSE, THREAD_STOPPED);
    } else {
        osthread_set_status(self, -1, THREAD_RUNNING);
    }
    return Qnil;
}

void *osthread_op_handler(int sig) {
    sigset_t set;

    // If we already stopped, we are running to resume, so just 
    // return so sigsuspend can go forward.
    if(GET_THREAD(STATE_VAR(thread))->status == THREAD_STOPPED) {
        return NULL;
    }
    if(GET_THREAD(STATE_VAR(thread))->status == THREAD_GC) {
        rb_gc_mark_stack(NULL, NULL);
        rb_gc_mark_threads();
    }

    /* If there is an error, we've been flagged to raise the 
     * exception. We set a exception_interrupt interrupt. */
    if(GET_THREAD(STATE_VAR(thread))->error) {
        STATE_VAR(exception_interrupt) = 1;
        return NULL;
    }

    sigfillset(&set);
    sigdelset(&set, SIG_RBOP);
    GET_THREAD(STATE_VAR(thread))->status = THREAD_STOPPED;
    sigsuspend(&set);
   
    GET_THREAD(STATE_VAR(thread))->status = THREAD_RUNNING;
    sched_yield();
    
    return NULL;
}

/* 
 * call-seq: Thread::OS.yield
 *
 * Yields the current Thread::OS from the processor to allow
 * another thread to run.
 */
VALUE rb_osthread_yield(VALUE self) {
    sched_yield();
    return Qnil;
}

const char * rb_thread_status_name(enum thread_status status);

/*
 * call-seq: th.status => String
 *
 * Returns a String which indicates the status of +th+
 */
VALUE rb_osthread_status(VALUE self) {
    return rb_str_new2(rb_thread_status_name(GET_THREAD(self)->status));
}

/*
 * call-seq: th.inspect => String
 *
 * Returns a String which describes +th+.
 */
VALUE rb_osthread_inspect(VALUE self) {
    VALUE str;

    str = rb_str_new(0, 11+7+16+9+1);
    sprintf(RSTRING(str)->ptr, "#<Thread::OS:0x%lx %s>", self, 
            rb_thread_status_name(GET_THREAD(self)->status));

    OBJ_INFECT(str, self);
    return str;
}

/*
 * call-seq: th.raise(exception)
 *
 * Raises an exception (see +Kernel::raise+) inside +th|
 */
VALUE rb_osthread_raise(int argc, VALUE *argv, VALUE self) {
    VALUE exc;

    exc = rb_make_exception(argc, argv);
    if(pthread_equal(pthread_self(), GET_THREAD(self)->os_thread)) {
        rb_raise_jump(exc);
    }

    GET_THREAD(self)->error = exc;
    rb_osthread_signal(self, SIG_RBOP);
    return Qnil;
}

void mutex_free(void *obj) {
    pthread_mutex_t *mutex;
    int state;
    mutex = (pthread_mutex_t*)obj;

    state = pthread_mutex_destroy(mutex);

    if(state == EBUSY) {
        /* Crap. it's locked guess we dont worry about it. */
    } else {
        free(mutex);
    }
    
}

/*
 * call-seq: Thread::OS::Mutex.new => mutex
 *
 * Creates a new OS mutex.
 */
VALUE rb_osthread_mutex_new(int argc, VALUE *argv, VALUE self) {
    VALUE obj;
    pthread_mutexattr_t *attr;
    pthread_mutex_t *mutex;

    /*

    if(!rb_scan_args(argc, argv, "01", &t)) {
        type = PTHREAD_MUTEX_DEFAULT;
    } else {
        type = NUM2INT(t);
    }

    */

    attr = (pthread_mutexattr_t*)malloc(sizeof(pthread_mutexattr_t));
    pthread_mutexattr_init(attr);
    //pthread_mutexattr_settype(attr, type);

    obj = Data_Make_Struct(self, pthread_mutex_t, 0, mutex_free, mutex);
    pthread_mutex_init(mutex, attr);

    return obj;
}

/*
 * call-seq: mutex.trylock
 *
 * Attempts to lock +mutex+. If the mutex is can not be locked, ie
 * it is already locked, +Thread::OS::Mutex::BusyMutex+ is raised.
 *
 * Upon success, true is returned.
 */
VALUE rb_osthread_mutex_trylock(VALUE self) {
    pthread_mutex_t *mutex;
    int state;
    Data_Get_Struct(self, pthread_mutex_t, mutex);

    state = pthread_mutex_trylock(mutex);

    switch(state) {
    case EBUSY:
        rb_raise(rb_path2class("Thread::OS::Mutex::BusyMutex"), "mutex locked");
    case EINVAL:
        rb_raise(rb_path2class("Thread::OS::Mutex::InvalidMutex"), "mutex not initialized");
    case 0:
    default:
        return Qtrue;
    }
}

/* 
 * call-seq: mutex.lock
 *
 * Locks +mutex+. If the mutex is already locked, the calling
 * thread blocks until the mutex is unlocked.
 */
VALUE rb_osthread_mutex_lock(VALUE self) {
    pthread_mutex_t *mutex;
    int state;
    VALUE err;
    Data_Get_Struct(self, pthread_mutex_t, mutex);

    state = pthread_mutex_lock(mutex);

    switch(state) {
    case EDEADLK:
        err = rb_const_get(CPATH(Thread::OS), _S(Deadlock));
        rb_raise(err, "locking would deadlock");
    case EINVAL:
        err = rb_const_get(CPATH(Thread::OS::Mutex), _S(InvalidMutex));
        rb_raise(err, "mutex not initialized");
    case 0:
    default:
        return Qtrue;
    }
}

/*
 * call-seq: mutex.unlock
 *
 * Unlocks +mutex+. If the mutex can not be unlocked
 * by the calling thread, +Thread::OS::InvalidPermission+ is raised.
 */
VALUE rb_osthread_mutex_unlock(VALUE self) {
    pthread_mutex_t *mutex;
    int state;
    VALUE err;
    Data_Get_Struct(self, pthread_mutex_t, mutex);

    state = pthread_mutex_unlock(mutex);

    switch(state) {
    case EPERM:
        err = rb_const_get(CPATH(Thread::OS), _S(InvalidPermission));
        rb_raise(err, "not owner of mutex");
    case EINVAL:
        err = rb_const_get(CPATH(Thread::OS::Mutex), _S(InvalidMutex));
        rb_raise(err, "mutex not initialized");
    case 0:
    default:
        return Qtrue;
    }
}

void Init_osthread() {
    struct rb_os_thread_t *oth;
    VALUE cOSThread, cOSThreadGroup, cOSMutex, error;

    cOSThread = rb_define_class_under(CPATH(Thread), "OS", rb_cObject);
    cOSThreadGroup = rb_define_class_under(CPATH(ThreadGroup), "OS", rb_cObject);
    error = rb_define_class_under(cOSThread, "Error", CPATH(Exception));
    rb_define_class_under(cOSThread, "UnknownThread", error);
    rb_define_class_under(cOSThread, "InvalidPermission", error);
    rb_define_class_under(cOSThread, "InvalidThread", error);
    rb_define_class_under(cOSThread, "DeadLock", error);
    rb_const_set(cOSThread, _S(SCHED_OTHER), INT2NUM(SCHED_OTHER));
    rb_const_set(cOSThread, _S(SCHED_RR), INT2NUM(SCHED_RR));
    rb_const_set(cOSThread, _S(SCHED_FIFO), INT2NUM(SCHED_FIFO));
    
    rb_const_set(cOSThreadGroup, _S(DEFAULT), 
            rb_class_new_instance(0, NULL, cOSThreadGroup));

    rb_define_method(cOSThread, "initialize", rb_osthread_init, 0);
    rb_define_method(cOSThread, "[]",  rb_osthread_aref, 1);
    rb_define_method(cOSThread, "[]=", rb_osthread_aset, 2);
    rb_define_method(cOSThread, "join", rb_osthread_join, 0);
    rb_define_method(cOSThread, "detach", rb_osthread_detach, 0);
    rb_define_method(cOSThread, "==", rb_osthread_equal, 1);
    rb_define_method(cOSThread, "policy", rb_osthread_policy, 0);
    rb_define_method(cOSThread, "policy=", rb_osthread_set_policy, 1);
    rb_define_method(cOSThread, "kill", rb_osthread_kill, 1);
    rb_define_method(cOSThread, "signal", rb_osthread_kill, 1);
    rb_define_method(cOSThread, "cancel", rb_osthread_cancel, 0);
    rb_define_method(cOSThread, "state", rb_osthread_state, 0);
    rb_define_method(cOSThread, "pause", rb_osthread_pause, 0);
    rb_define_method(cOSThread, "inspect", rb_osthread_inspect, 0);
    rb_define_method(cOSThread, "status", rb_osthread_status, 0);
    rb_define_method(cOSThread, "raise", rb_osthread_raise, -1);

    rb_define_singleton_method(cOSThread, "start", rb_osthread_start, -1);
    rb_define_singleton_method(cOSThread, "new", rb_osthread_start, -1);
    rb_define_singleton_method(cOSThread, "fork", rb_osthread_start, -1);
    rb_define_singleton_method(cOSThread, "thread_id", rb_osthread_id, 0);
    rb_define_singleton_method(cOSThread, "current", rb_osthread_current, 0);
    rb_define_singleton_method(cOSThread, "list", rb_osthread_list, 0);
    rb_define_singleton_method(cOSThread, "yield", rb_osthread_yield, 0);
    rb_define_singleton_method(cOSThread, "pass", rb_osthread_yield, 0);

    STATE_VAR(thread) = 
        Data_Make_Struct(CPATH(Thread::OS), struct rb_os_thread_t, 0, 0, oth);

    oth->os_thread = pthread_self();
    oth->locals = rb_hash_new();
    oth->state = rb_get_state();
    oth->status = THREAD_RUNNING;
    STATE_VAR(os_threads) = rb_ary_new();
    rb_global_variable(&STATE_VAR(os_threads));
    rb_ary_push(STATE_VAR(os_threads), STATE_VAR(thread));

    signal(SIG_RBOP, (void*)osthread_op_handler);

    cOSMutex = rb_define_class_under(cOSThread, "Mutex", rb_cObject);
    rb_define_singleton_method(cOSMutex, "new", rb_osthread_mutex_new, -1);
    rb_define_method(cOSMutex, "trylock", rb_osthread_mutex_trylock, 0);
    rb_define_method(cOSMutex, "lock", rb_osthread_mutex_lock, 0);
    rb_define_method(cOSMutex, "unlock", rb_osthread_mutex_unlock, 0);
    /*
    rb_const_set(cOSMutex, "NORMAL", INT2NUM(PTHREAD_MUTEX_NORMAL));
    rb_const_set(cOSMutex, "RECURSIVE", INT2NUM(PTHREAD_MUTEX_RECURSIVE));
    rb_const_set(cOSMutex, "ERRORCHECK", INT2NUM(PTHREAD_MUTEX_ERRORCHECK));
    rb_const_set(cOSMutex, "DEFAULT", INT2NUM(PTHREAD_MUTEX_DEFAULT));
    */

    rb_define_class_under(cOSMutex, "BusyMutex", error);
    rb_define_class_under(cOSMutex, "InvalidMutex", error);
    
    //cOSThreadAttr = rb_define_class_under(CPATH(Thread::OS), "Attributes", rb_cObject);
    //rb_define_method(cOSThreadAttr, "initialize", rb_osthread_attr_init, 0);
}

