require 'backtrace'

require 'test/unit'

class TestRubyBacktrace < Test::Unit::TestCase
    def get_backtrace
        begin
            raise "blah!"
        rescue Exception => e
            return e.backtrace
        end
    end

    def test_backtrace
        assert_kind_of(Ruby::Backtrace, get_backtrace)
    end

    def test_file
        bt = get_backtrace
        assert_equal(__FILE__, bt.file)
    end

    def test_line
        bt = get_backtrace
        assert_equal(8, bt.line)
    end

    def test_current_method
        bt = get_backtrace
        assert_equal(:get_backtrace, bt.current_method)
    end

    def test_frames
        bt = get_backtrace
        assert_kind_of(Array, bt.frames)
        assert_kind_of(Ruby::Frame, bt.frames.first)
    end

    def test_location
        bt = get_backtrace
        assert_equal("#{__FILE__}:8 in 'get_backtrace'", bt.location)
    end

    def test_print
        bt = get_backtrace
        str = bt.print
        assert_match(/'test_print'/m, str)
    end

    def test_print_truncate
        bt = get_backtrace

        # This could be more completed, but i dont want it to depend
        # on test/unit's call path, so just test the head count.
        str = bt.print(1,1,1)
        assert_match(/.*?\n\t .../m,str)
    end

    def test_contains_position
        bt = get_backtrace
        assert(bt.contains_position?)

        bt = Ruby::Backtrace.new
        assert(!bt.contains_position?)
    end

    def test_error_handler
        IO.popen("-") do |io|
            if io
                str = io.read
                assert_match(/from error handler/,str)
            else
                Ruby.error_handler do
                    STDOUT.puts "from error handler"
                end

                # An icky little hack to break Test::Unit::TestCase
                # in the child so the error hits the top of the stack.
                class << self
                    def add_error(x)
                        raise x
                    end
                end

                raise "here we go! #{Process.pid}"
            end
        end
    end
end
