
require 'backtrace'

require 'test/unit'

class TestRubyFrame < Test::Unit::TestCase

    def test_current
        fr = Ruby::Frame.current
        assert_kind_of(Ruby::Frame, fr)
    end

    def test_position
        fr = Ruby::Frame.current  # adjust the number below to reflect the current line!
        assert_equal(14, fr.line) 
        assert_equal(__FILE__, fr.file)
        assert_equal([__FILE__, 14], fr.position)
    end

    def test_current_method
        fr = jump!
        assert_equal(:jump!, fr.current_method)
    end

    def jump!
        Ruby::Frame.current
    end

    def test_self_and_class
        q = Q.new
        fr = q.go
        assert_equal(q, fr.self)
        assert_equal(Q, fr.current_class)
    end

    class Q
        def go
            Ruby::Frame.current
        end
    end

    def test_nice_string
        fr = jump!
        assert_equal("#{__FILE__}:26:in 'jump!'", fr.nice_string)
    end

    def test_metadata
        assert_kind_of(Hash, Ruby::Frame.current.metadata)
    end

    def test_metadata_scope
        Ruby::Frame.current.metadata[:name] = "blah"

        assert_equal("blah", get_name)
    end

    def get_name
        caller.frames.first.metadata[:name]
    end
end


